import sys
import PyQt5.QtWidgets as QtWidgets
import PyQt5.QtCore as QtCore
import PyQt5.QtGui as QtGui
import PyQt5.uic as uic
import backend
import enum
import os


class MSG_TYPE(enum.Enum):
    CONTROL = 0
    SERIAL = 1
    SEQUENCE = 2
    CAMERA = 3


class MainWindow(QtWidgets.QMainWindow):
    msg_lookup = {MSG_TYPE.CONTROL: "m_control",
                  MSG_TYPE.SERIAL: "m_serial",
                  MSG_TYPE.CAMERA: "m_camera",
                  MSG_TYPE.SEQUENCE: "m_sequence"}
    timer_msec = 4

    def __init__(self, flags, ui_path):
        self.pixmap = None
        super(MainWindow, self).__init__(flags=flags)
        uic.loadUi(ui_path, self)

        self.g_control = self.findChild(QtCore.QObject, "group_control")
        self.g_camera = self.findChild(QtCore.QObject, "group_camera")
        self.g_serial = self.findChild(QtCore.QObject, "group_serial")
        self.g_sequence = self.findChild(QtCore.QObject, "group_sequence")

        self.rail_server = backend.MovingRailServer(self)
        self.camera_server = backend.CameraServer(self)
        # timer to refresh content sometimes
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.on_timer)

        self.steps_todo = []
        self.sequence_i = -1
        self.sequence_t = 0.0
        self.sequence_params = None
        self.is_shooting = False

        self.init_control()
        self.init_camera()
        self.init_serial()
        self.init_sequence()

        # Need to connect to serial AND perform homing
        self.toggle_control(False)
        self.toggle_sequence(False)

        self.timer.start(self.timer_msec)
        self.show()

    def init_control(self):
        self.display_msg("", MSG_TYPE.CONTROL)

        b = self.findChild(QtWidgets.QPushButton, "b_homing")
        b.clicked.connect(self.on_homing)

        b = self.findChild(QtWidgets.QPushButton, "b_motor_off")
        b.clicked.connect(lambda: self.on_motor_change(False))

        b = self.findChild(QtWidgets.QPushButton, "b_motor_on")
        b.clicked.connect(lambda: self.on_motor_change(True))

        b = self.findChild(QtWidgets.QSlider, "s_target")
        b.valueChanged.connect(self.on_slider_update)

        b = self.findChild(QtWidgets.QLineEdit, "i_target")
        b.textChanged.connect(self.on_target_update)

        b = self.findChild(QtWidgets.QLineEdit, "o_current")
        b.textChanged.connect(self.on_current_update)

        b = self.findChild(QtWidgets.QPushButton, "b_go")
        b.clicked.connect(self.on_go)

    def display_msg(self, txt, msg_type):
        m_name = self.msg_lookup[msg_type]
        m = self.findChild(QtWidgets.QLabel, m_name)
        m.setText(txt)

    def init_camera(self):
        self.display_msg("", MSG_TYPE.CAMERA)

        b = self.findChild(QtWidgets.QPushButton, "b_refresh")
        b.clicked.connect(self.on_refresh_image)

        b = self.findChild(QtWidgets.QPushButton, "b_shoot")
        b.clicked.connect(self.on_shoot)

        b = self.findChild(QtWidgets.QPushButton, "b_browse")
        b.clicked.connect(self.on_browse)

        b = self.findChild(QtWidgets.QLineEdit, "i_output_folder")
        b.textChanged.connect(self.on_folder_change)
        b.setText("./output")

    def init_serial(self):
        self.display_msg("No active serial!", MSG_TYPE.SERIAL)

        i_port = self.findChild(QtWidgets.QComboBox, "i_port")
        devices = self.rail_server.scan_devices()
        i_port.clear()
        i_port.addItems(devices)

        i_baudrate = self.findChild(QtWidgets.QComboBox, "i_baudrate")
        i_baudrate.setCurrentText("9600")

        b_connect = self.findChild(QtWidgets.QPushButton, "b_connect")
        b_connect.clicked.connect(self.on_connect)

    def on_motor_change(self, state):
        success = self.rail_server.motor_enabled(state)
        if success and not state:
            self.toggle_control(False)
        msg = "Motors on" if state else "Motors off"
        self.display_msg(msg, MSG_TYPE.CONTROL)

    def init_sequence(self):
        b = self.findChild(QtWidgets.QPushButton, "b_step_start")
        b.clicked.connect(self.on_start_sequence)

        b = self.findChild(QtWidgets.QPushButton, "b_step_cancel")
        b.clicked.connect(self.on_cancel_sequence)

        b = self.findChild(QtWidgets.QCheckBox, "i_step_shoot")
        b.stateChanged.connect(self.on_change_sequence)
        for child in self.g_sequence.findChildren(QtWidgets.QLineEdit):
            child.textChanged.connect(self.on_change_sequence)
        for child in self.g_sequence.findChildren(QtWidgets.QAbstractSpinBox):
            child.valueChanged.connect(self.on_change_sequence)

        self.on_change_sequence()

    def on_connect(self):
        b_connect = self.findChild(QtWidgets.QPushButton, "b_connect")
        i_port = self.findChild(QtWidgets.QComboBox, "i_port")
        i_baudrate = self.findChild(QtWidgets.QComboBox, "i_baudrate")
        text = b_connect.text()
        if text == "Connect":
            device = i_port.currentText()
            baudrate = int(i_baudrate.currentText())
            success = self.rail_server.connect(device, baudrate)
            if success:
                i_port.setEnabled(False)
                i_baudrate.setEnabled(False)
                self.display_msg("Connected to: "+device, MSG_TYPE.SERIAL)
                b_connect.setText("Disconnect")
        else:
            self.rail_server.close()
            self.toggle_control(False)
            self.toggle_sequence(False)
            i_port.setEnabled(True)
            i_baudrate.setEnabled(True)
            self.display_msg("No active serial!", MSG_TYPE.SERIAL)
            b_connect.setText("Connect")

    def on_timer(self):
        self.rail_server.update_read()
        in_sequence = self.sequence_i > -1
        shooting_done = False
        if self.is_shooting:
            shooting_done = self.camera_server.get_progress()
            if shooting_done:
                self.is_shooting = False
                self.on_refresh_image()

        if in_sequence:
            self.update_sequence_text()
            shoot_enabled = self.sequence_params.shoot
            i_p, i_t = self.sequence_params[self.sequence_i]
            self.sequence_t += self.timer_msec / 1000.0
            if self.sequence_t >= i_t:
                if shoot_enabled:
                    if shooting_done:
                        self.move_next()
                    elif not self.is_shooting:
                        self.on_shoot()
                else:
                    self.move_next()

    def on_refresh_image(self):
        filename = self.camera_server.get_last_img()
        if filename is None:
            self.display_msg("No image found", MSG_TYPE.CAMERA)
        else:
            b = self.findChild(QtWidgets.QLabel, "image_viewer")
            self.pixmap = QtGui.QPixmap(filename)
            pixmap_scaled = self.pixmap.scaled(b.size(), QtCore.Qt.KeepAspectRatio)
            b.setPixmap(pixmap_scaled)

    def move_next(self):
        self.sequence_i += 1
        if self.sequence_i == self.sequence_params.n_steps:
            self.on_cancel_sequence()
            return
        i_p, i_t = self.sequence_params[self.sequence_i]
        self.move(i_p, speed=self.rail_server.max_speed)

    def move(self, pos, speed):
        m = self.findChild(QtWidgets.QLineEdit, "o_current")
        success = self.rail_server.move(pos, speed)
        if success:
            m.setText("{:.1f}".format(pos))
            self.display_msg("Moved to: {}".format(pos), MSG_TYPE.CONTROL)
        else:
            self.display_msg("Error: move failed.", MSG_TYPE.CONTROL)

    def on_go(self):
        i_t = self.findChild(QtWidgets.QLineEdit, "i_target")
        i_s = self.findChild(QtWidgets.QLineEdit, "i_speed")
        try:
            pos = float(i_t.text())
            speed = float(i_s.text())
            if speed <= 0:
                raise ValueError
            self.move(pos, speed)
        except ValueError:
            self.display_msg("Error: incorrect target.", MSG_TYPE.CONTROL)

    def on_shoot(self):
        self.camera_server.shoot()
        self.is_shooting = True

    def resizeEvent(self, *args, **kwargs):
        if self.pixmap is not None:
            b = self.findChild(QtWidgets.QLabel, "image_viewer")
            pixmap_scaled = self.pixmap.scaled(b.size(), QtCore.Qt.KeepAspectRatio)
            b.setPixmap(pixmap_scaled)
        super(QtWidgets.QMainWindow, self).resizeEvent(*args, **kwargs)

    def on_homing(self):
        success = self.rail_server.home()
        if success:
            s1 = self.findChild(QtWidgets.QLineEdit, "i_target")
            s2 = self.findChild(QtWidgets.QLineEdit, "o_current")
            s1.setText("0.0")
            s2.setText("0.0")
            self.toggle_control(True)
            self.toggle_sequence(True)

    def toggle_control(self, state):
        for child in self.g_control.findChildren(QtWidgets.QWidget):
            if child.objectName() in ("b_homing", "b_motor_on", "s_current", "o_current"):
                continue
            child.setEnabled(state)

    def toggle_serial(self, state):
        for child in self.g_serial.findChildren(QtWidgets.QWidget):
            child.setEnabled(state)

    def toggle_sequence(self, state):
        for child in self.g_sequence.findChildren(QtWidgets.QWidget):
            if child.objectName() in ("b_step_cancel", "i_step_start"):
                continue
            child.setEnabled(state)

    def toggle_camera(self, state):
        for child in self.g_camera.findChildren(QtWidgets.QWidget):
            if child.objectName() in ("b_refresh", "image_viewer"):
                continue
            child.setEnabled(state)

    def on_slider_update(self):
        s = self.findChild(QtWidgets.QSlider, "s_target")
        b = self.findChild(QtWidgets.QLineEdit, "i_target")
        val = s.value()
        b.setText(str(val))

    def on_target_update(self):
        s = self.findChild(QtWidgets.QSlider, "s_target")
        i = self.findChild(QtWidgets.QLineEdit, "i_target")
        try:
            val = float(i.text())
            if s.minimum() <= val <= s.maximum():
                s.setValue(int(val))
        except ValueError:
            pass

    def on_current_update(self):
        i = self.findChild(QtWidgets.QLineEdit, "o_current")
        s = self.findChild(QtWidgets.QSlider, "s_current")
        m_step = self.findChild(QtWidgets.QLineEdit, "i_step_start")
        val = float(i.text())
        s.setValue(int(val))
        if self.sequence_i == -1:
            m_step.setText(i.text())

    def on_start_sequence(self):
        self.toggle_sequence(False)
        self.toggle_control(False)
        self.toggle_serial(False)
        self.toggle_camera(False)
        b = self.findChild(QtWidgets.QLabel, "m_sequence")
        b.setEnabled(True)
        b = self.findChild(QtWidgets.QPushButton, "b_homing")
        b.setEnabled(False)
        b = self.findChild(QtWidgets.QPushButton, "b_motor_on")
        b.setEnabled(False)
        self.sequence_t = 0.0
        self.sequence_i = 0

    def on_cancel_sequence(self):
        b = self.findChild(QtWidgets.QPushButton, "b_homing")
        b.setEnabled(True)
        b = self.findChild(QtWidgets.QPushButton, "b_motor_on")
        b.setEnabled(True)
        self.toggle_sequence(True)
        self.toggle_control(True)
        self.toggle_serial(True)
        self.toggle_camera(True)
        self.sequence_params = None
        self.sequence_t = 0.0
        self.sequence_i = -1
        self.on_current_update()

    def on_change_sequence(self):
        i_h = self.findChild(QtWidgets.QSpinBox, "i_hour")
        i_m = self.findChild(QtWidgets.QSpinBox, "i_minute")
        i_s = self.findChild(QtWidgets.QDoubleSpinBox, "i_second")
        i_start = self.findChild(QtWidgets.QWidget, "i_step_start")
        i_end = self.findChild(QtWidgets.QWidget, "i_step_end")
        i_n = self.findChild(QtWidgets.QWidget, "i_step_n")
        i_shoot = self.findChild(QtWidgets.QCheckBox, "i_step_shoot")
        try:
            dt = i_h.value()*(24*60)+i_m.value()*60+i_s.value()
            start = float(i_start.text())
            end = float(i_end.text())
            n = int(i_n.text())
            if dt <= 0 or n <= 1:
                raise ValueError
            shoot = i_shoot.isChecked()
            self.sequence_params = SequenceParams(start, end, n, dt, shoot)
            self.update_sequence_text()
        except ValueError:
            self.display_msg("Error: invalid settings.", MSG_TYPE.SEQUENCE)
            return

    def update_sequence_text(self):
        p = self.sequence_params
        if p is None:
            return
        t1 = self.sequence_t
        _, t2 = p[-1]
        t1_txt = self._time_str(t1)
        t2_txt = self._time_str(t2)
        i1 = max(0, self.sequence_i)
        i2 = p.n_steps
        self.display_msg("step {}/{}, time {}/{}".format(i1, i2, t1_txt, t2_txt), MSG_TYPE.SEQUENCE)

    def _time_str(self, s):
        h = int(s//3600)
        s -= h*3600
        m = int(s//60)
        s -= m*60
        s = int(s)
        return "{:02d}:{:02d}:{:02d}".format(h, m, s)

    def append_gcode(self, txt):
        m = self.findChild(QtWidgets.QTextBrowser, "o_gcode")
        m.append(txt)

    def on_folder_change(self):
        b = self.findChild(QtWidgets.QLineEdit, "i_output_folder")
        folder = b.text()
        self.camera_server.set_folder(folder)

    def on_browse(self):
        d = QtWidgets.QFileDialog()
        options = d.options()
        folder = d.getExistingDirectory(None, "Choose an output directory", ".", options=options)
        b = self.findChild(QtWidgets.QLineEdit, "i_output_folder")
        b.setText(folder)

    def close(self):
        self.camera_server.close()
        self.rail_server.close()


class SequenceParams:
    def __init__(self, start, end, n, dt, shoot=True):
        step = (end-start)/(n-1)
        self.list_pos = [start+i*step for i in range(n)]
        self.list_time = [i * dt for i in range(n)]
        self.n_steps = n
        self.shoot = shoot

    def __getitem__(self, item):
        return self.list_pos[item], self.list_time[item]


def launch_app(folder):
    app = QtWidgets.QApplication(sys.argv)
    flags = QtCore.Qt.WindowFlags()
    ui_path = os.path.join(folder, "QtDesigner/design.ui")
    window = MainWindow(flags, ui_path)
    app.exec()
